"""Extract spectral informations from data."""
import numpy as np
from joblib import Parallel, delayed
from scipy.signal import hilbert, filtfilt
from scipy import fftpack

from tensorpac.config import CONFIG


def hilbertm(x):
    """Faster Hilbert fix.

    x must have a shape of (..., n_pts)
    """
    n_pts = x.shape[-1]
    fc = fftpack.helper.next_fast_len(n_pts)
    return hilbert(x, fc, axis=-1)[..., 0:n_pts]


def spectral(x, sf, f, stype, dcomplex, cycle, width, n_jobs):
    """Extract spectral informations from data.

    Parameters
    ----------
    x : array_like
        Array of data
    sf : float
        Sampling frequency
    f : array_like
        Frequency vector of shape (N, 2)
    stype : string
        Spectral informations to extract (use either 'pha' or 'amp')
    dcomplex : string
        Complex decomposition type. Use either 'hilbert' or 'wavelet'
    cycle : int
        Number of cycles to use for fir1 filtering.
    width : int
        Width of the wavelet.
    n_jobs : int
        Number of jobs to use. If jobs is -1, all of them are going to be
        used.
    """
    n_freqs = f.shape[0]
    # Filtering + complex decomposition :
    if dcomplex is 'hilbert':
        # get filtering coefficients
        b = []
        a = np.zeros((n_freqs,), dtype=float)
        forder = np.zeros((n_freqs,), dtype=int)
        for k in range(n_freqs):
            forder[k] = fir_order(sf, x.shape[-1], f[k, 0], cycle=cycle)
            _b, a[k] = fir1(forder[k], f[k, :] / (sf / 2.))
            b += [_b]
        # Filt each time series :
        xf = Parallel(n_jobs=n_jobs, **CONFIG['JOBLIB_CFG'])(delayed(filtfilt)(
            b[k], a[k], x, padlen=forder[k], axis=-1) for k in range(n_freqs))
        # Use hilbert for the complex decomposition :
        xd = np.asarray(xf)
        if stype is not None:
            xd = hilbertm(xd)
    elif dcomplex is 'wavelet':
        f = f.mean(1)  # centered frequencies
        xd = Parallel(n_jobs=n_jobs, **CONFIG['JOBLIB_CFG'])(delayed(morlet)(
            x, sf, k, width) for k in f)

    # Extract phase / amplitude :
    if stype is 'pha':
        return np.angle(xd).astype(np.float64)
    elif stype is 'amp':
        return np.abs(xd).astype(np.float64)
    elif stype is None:
        return xd.astype(np.float64)

###############################################################################
###############################################################################
#                            FILTERING
###############################################################################
###############################################################################


def fir_order(fs, sizevec, flow, cycle=3):
    filtorder = cycle * (fs // flow)

    if (sizevec < 3 * filtorder):
        filtorder = (sizevec - 1) // 3

    return int(filtorder)


def n_odd_fcn(f, o, w, l):
    """Odd case."""
    # Variables :
    b0 = 0
    m = np.array(range(int(l + 1)))
    k = m[1:len(m)]
    b = np.zeros(k.shape)

    # Run Loop :
    for s in range(0, len(f), 2):
        m = (o[s + 1] - o[s]) / (f[s + 1] - f[s])
        b1 = o[s] - m * f[s]
        b0 = b0 + (b1 * (f[s + 1] - f[s]) + m / 2 * (
            f[s + 1] * f[s + 1] - f[s] * f[s])) * abs(
            np.square(w[round((s + 1) / 2)]))
        b = b + (m / (4 * np.pi * np.pi) * (
            np.cos(2 * np.pi * k * f[s + 1]) - np.cos(2 * np.pi * k * f[s])
        ) / (k * k)) * abs(np.square(w[round((s + 1) / 2)]))
        b = b + (f[s + 1] * (m * f[s + 1] + b1) * np.sinc(2 * k * f[
            s + 1]) - f[s] * (m * f[s] + b1) * np.sinc(2 * k * f[s])) * abs(
            np.square(w[round((s + 1) / 2)]))

    b = np.insert(b, 0, b0)
    a = (np.square(w[0])) * 4 * b
    a[0] = a[0] / 2
    aud = np.flipud(a[1:len(a)]) / 2
    a2 = np.insert(aud, len(aud), a[0])
    h = np.concatenate((a2, a[1:] / 2))

    return h


def n_even_fcn(f, o, w, l):
    """Even case."""
    # Variables :
    k = np.array(range(0, int(l) + 1, 1)) + 0.5
    b = np.zeros(k.shape)

    # # Run Loop :
    for s in range(0, len(f), 2):
        m = (o[s + 1] - o[s]) / (f[s + 1] - f[s])
        b1 = o[s] - m * f[s]
        b = b + (m / (4 * np.pi * np.pi) * (np.cos(2 * np.pi * k * f[
            s + 1]) - np.cos(2 * np.pi * k * f[s])) / (
            k * k)) * abs(np.square(w[round((s + 1) / 2)]))
        b = b + (f[s + 1] * (m * f[s + 1] + b1) * np.sinc(2 * k * f[
            s + 1]) - f[s] * (m * f[s] + b1) * np.sinc(2 * k * f[s])) * abs(
            np.square(w[round((s + 1) / 2)]))

    a = (np.square(w[0])) * 4 * b
    h = 0.5 * np.concatenate((np.flipud(a), a))

    return h


def firls(n, f, o):
    # Variables definition :
    w = np.ones(round(len(f) / 2))
    n += 1
    f /= 2
    lo = (n - 1) / 2

    nodd = bool(n % 2)

    if nodd:  # Odd case
        h = n_odd_fcn(f, o, w, lo)
    else:  # Even case
        h = n_even_fcn(f, o, w, lo)

    return h


def fir1(n, wn):
    # Variables definition :
    nbands = len(wn) + 1
    ff = np.array((0, wn[0], wn[0], wn[1], wn[1], 1))

    f0 = np.mean(ff[2:4])
    lo = n + 1

    mags = np.array(range(nbands)).reshape(1, -1) % 2
    aa = np.ravel(np.tile(mags, (2, 1)), order='F')

    # Get filter coefficients :
    h = firls(lo - 1, ff, aa)

    # Apply a window to coefficients :
    wind = np.hamming(lo)
    b = h * wind
    c = np.exp(-1j * 2 * np.pi * (f0 / 2) * np.array(range(lo)))
    b /= abs(c @ b)

    return b, 1


###############################################################################
###############################################################################
#                            FILTERING
###############################################################################
###############################################################################


def morlet(x, sf, f, width=7.):
    """Complex decomposition of a signal x using the morlet wavelet.

    Parameters
    ----------
    x : array_like, shape (N,)
        The signal to use for the complex decomposition. Must be
        a vector of length N.
    sf : float
        Sampling frequency
    f : array_like, shape (2,)
        Frequency vector
    width : float | 7.
        Width of the wavelet

    Returns
    -------
    xout: array_like, shape (N,)
        The complex decomposition of the signal x.
    """
    dt = 1 / sf
    sf = f / width
    st = 1 / (2 * np.pi * sf)

    # Build morlet's wavelet :
    t = np.arange(-width * st / 2, width * st / 2, dt)
    a = 1 / np.sqrt((st * np.sqrt(np.pi)))
    m = a * np.exp(-np.square(t) / (2 * np.square(st))) * np.exp(
        1j * 2 * np.pi * f * t)

    def ndmorlet(xt):
        # Compute morlet :
        y = np.convolve(xt, m)
        return y[int(np.ceil(len(m) / 2)) - 1:int(len(y) - np.floor(
            len(m) / 2))]
    return np.apply_along_axis(ndmorlet, -1, x)



"""Individual methods for assessing ERPAC."""
import numpy as np
from scipy.stats import chi2

from joblib import Parallel, delayed

from tensorpac.gcmi import nd_mi_gg
from tensorpac.config import CONFIG


def pearson(x, y, st='i...j, k...j->ik...'):
    """Pearson correlation for multi-dimensional arrays.

    Parameters
    ----------
    x, y : array_like
        Compute pearson correlation between the multi-dimensional arrays
        x and y.
    st : string | 'i..j, k..j->ik...'
        The string to pass to the np.einsum function.

    Returns
    -------
    cov: array_like
        The pearson correlation array.
    """
    n = x.shape[-1]
    # Distribution center :
    mu_x = x.mean(-1, keepdims=True)
    mu_y = y.mean(-1, keepdims=True)
    # Distribution deviation :
    s_x = x.std(-1, ddof=n - 1, keepdims=True)
    s_y = y.std(-1, ddof=n - 1, keepdims=True)
    # Compute correlation coefficient :
    cov = np.einsum(st, x, y)
    mu_xy = np.einsum(st, mu_x, mu_y)
    cov -= n * mu_xy
    cov /= np.einsum(st, s_x, s_y)
    return cov


def erpac(pha, amp):
    """Event-Related Phase Amplitude Coupling.

    This function computed the correlation coefficient between a circular and a
    linear random variable at each time point and across trials. Adapted from
    the function circ_corrcc Circular Statistics Toolbox for
    Matlab By Philipp Berens, 2009 :cite:`berens2009circstat`. This function is
    an adaptation of Voytek, 2013 :cite:`voytek2013method` for tensors.

    Parameters
    ----------
    pha, amp : array_like
        Respectively the arrays of phases of shape (n_pha, ..., n_epochs) and
        the array of amplitudes of shape (n_amp, ..., n_epochs).

    Returns
    -------
    rho : array_like
        Array of correlation coefficients of shape (n_amp, n_pha, ...)
    pval : array_like
        Array of p-values of shape (n_amp, n_pha, ...).

    References
    ----------
    Voytek et al. 2013 :cite:`voytek2013method`
    """
    # Compute correlation coefficient for sin and cos independently
    n = pha.shape[-1]
    sa, ca = np.sin(pha), np.cos(pha)
    rxs = pearson(amp, sa)
    rxc = pearson(amp, ca)
    rcs = pearson(sa, ca, st='i...j, k...j->i...')
    rcs = rcs[np.newaxis, ...]

    # Compute angular-linear correlation (equ. 27.47)
    rho = np.sqrt((rxc**2 + rxs**2 - 2 * rxc * rxs * rcs) / (1 - rcs**2))

    # Compute pvalue :
    pval = 1. - chi2.cdf(n * rho**2, 2)

    return rho, pval


def ergcpac(pha, amp, smooth=None, n_jobs=-1):
    """Event Related PAC using the Gaussian Copula Mutual Information.

    This function assumes that phases and amplitudes have already been
    prepared i.e. phases should be represented in a unit circle
    (np.c_[np.sin(pha), np.cos(pha)]) and both inputs should also have been
    copnormed.

    Parameters
    ----------
    pha, amp : array_like
        Respectively arrays of phases of shape (n_pha, n_times, 2, n_epochs)
        and the array of amplitudes of shape (n_amp, n_times, 1, n_epochs).

    Returns
    -------
    erpac : array_like
        Array of correlation coefficients of shape (n_amp, n_pha, n_times)

    References
    ----------
    Ince et al. 2017 :cite:`ince2017statistical`
    """
    # get shapes
    (n_pha, n_times, _, n_epochs), n_amp = pha.shape, amp.shape[0]  # noqa
    # compute mutual information across trials
    ergcpac = np.zeros((n_amp, n_pha, n_times))
    if isinstance(smooth, int):
        # define the temporal smoothing vector
        vec = np.arange(smooth, n_times - smooth, 1)
        times = [slice(k - smooth, k + smooth + 1) for k in vec]
        # move time axis to avoid to do it inside parallel
        pha, amp = np.moveaxis(pha, 1, -2), np.moveaxis(amp, 1, -2)
        # function to run in parallel across times
        def _fcn(t):  # noqa
            _erpac = np.zeros((n_amp, n_pha), dtype=float)
            xp, xa = pha[..., t, :], amp[..., t, :]
            for a in range(n_amp):
                _xa = xa.reshape(n_amp, 1, -1)
                for p in range(n_pha):
                    _xp = xp.reshape(n_pha, 2, -1)
                    _erpac[a, p] = nd_mi_gg(_xp[p, ...], _xa[a, ...])
            return _erpac
        # run the function across time points
        _ergcpac = Parallel(n_jobs=n_jobs, **CONFIG['JOBLIB_CFG'])(delayed(
            _fcn)(t) for t in times)
        # reconstruct the smoothed array
        for a in range(n_amp):
            for p in range(n_pha):
                mean_vec = np.zeros((n_times,), dtype=float)
                for t, _gc in zip(times, _ergcpac):
                    ergcpac[a, p, t] += _gc[a, p]
                    mean_vec[t] += 1
                ergcpac[a, p, :] /= mean_vec
    else:
        for a in range(n_amp):
            for p in range(n_pha):
                ergcpac[a, p, ...] = nd_mi_gg(pha[p, ...], amp[a, ...])
    return ergcpac


def swap_erpac_trials(pha):
    """Swap trials across the last dimension."""
    tr_ = np.random.permutation(pha.shape[-1])
    return pha[..., tr_]


def _ergcpac_perm(pha, amp, smooth=None, n_jobs=-1, n_perm=200):
    def _ergcpac_single_perm():
        p = swap_erpac_trials(pha)
        return ergcpac(p, amp, smooth=smooth, n_jobs=1)
    out = Parallel(n_jobs=n_jobs, **CONFIG['JOBLIB_CFG'])(delayed(
        _ergcpac_single_perm)() for _ in range(n_perm))
    return np.stack(out)



"""Tensorpac methods."""
from .meth_pac import (
    mean_vector_length, modulation_index, heights_ratio, norm_direct_pac,
    phase_locking_value, gauss_cop_pac)  # noqa
from .meth_pac_nb import (
    mean_vector_length_nb, modulation_index_nb, heights_ratio_nb,
    norm_direct_pac_nb, phase_locking_value_nb)  # noqa
from .meth_erpac import (erpac, ergcpac, _ergcpac_perm)  # noqa
from .meth_pp import (preferred_phase)  # noqa
from .meth_surrogates import (compute_surrogates, swap_pha_amp, swap_blocks,  # noqa
                              time_lag, normalize)
from .meth_switch import (pacstr, get_pac_fcn)  # noqa



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# License: 3-clause BSD
import os
from setuptools import setup, find_packages

__version__ = "0.6.4"
NAME = 'tensorpac'
AUTHOR = "Etienne Combrisson"
MAINTAINER = "Etienne Combrisson"
EMAIL = 'e.combrisson@gmail.com'
KEYWORDS = "phase-amplitude coupling pac tensor oscillation meg eeg python"
DESCRIPTION = "Tensor-based Phase-Amplitude Coupling"
URL = 'http://etiennecmb.github.io/tensorpac/'
DOWNLOAD_URL = ("https://github.com/EtienneCmb/tensorpac/archive/v" +
                __version__ + ".tar.gz")
# Data path :
PACKAGE_DATA = {}


def read(fname):
    """Read README and LICENSE."""
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name=NAME,
    version=__version__,
    packages=find_packages(),
    package_dir={'tensorpac': 'tensorpac'},
    package_data=PACKAGE_DATA,
    include_package_data=True,
    description=DESCRIPTION,
    long_description=read('README.rst'),
    platforms='any',
    setup_requires=['numpy', 'joblib'],
    install_requires=[
        "numpy",
        "scipy",
        "joblib"
    ],
    dependency_links=[],
    author=AUTHOR,
    maintainer=MAINTAINER,
    author_email=EMAIL,
    url=URL,
    download_url=DOWNLOAD_URL,
    license="BSD 3-Clause License",
    keywords=KEYWORDS,
    classifiers=["Development Status :: 5 - Production/Stable",
                 'Intended Audience :: Science/Research',
                 'Intended Audience :: Education',
                 'Intended Audience :: Developers',
                 "Programming Language :: Python :: 3.6",
                 "Programming Language :: Python :: 3.7",
                 "Programming Language :: Python :: 3.7"
                 ])



"""Test random signals generation."""
import numpy as np

from tensorpac.signals import pac_signals_wavelet, pac_signals_tort


class TestSignals(object):
    """Test random signals generation."""

    def test_pac_signals_dtrials(self):
        """Definition of artificially coupled signals using dPha/dAmp."""
        assert pac_signals_tort(f_pha=5, f_amp=130, sf=512, n_epochs=23,
                                chi=0.9, noise=2, dpha=35, damp=46)

    def test_pac_signals_bandwidth(self):
        """Definition of artificially coupled signals using bandwidth."""
        assert pac_signals_tort(f_pha=[5, 7], f_amp=[30, 60], sf=200.,
                                n_epochs=100, chi=0.5, noise=3., n_times=1000)
        assert pac_signals_wavelet(f_pha=10, f_amp=57., n_times=1240, sf=256,
                                   noise=.7, n_epochs=33, pp=np.pi / 4,
                                   rnd_state=23)

    def test_default_args(self):
        """Test default aurguments for pac_vec."""
        assert pac_signals_tort(chi=2., noise=11., dpha=120., damp=200.)



"""Tensorpac configuration."""

# Gaussian-Copula configuration
MI_BIASCORRECT = False
MI_DEMEAN = False

# Joblib config
JOBLIB_CFG = dict()  # prefer='threads'

# MNE config
try:
    import mne
    MNE_EPOCHS_TYPE = [mne.Epochs, mne.EpochsArray, mne.epochs.BaseEpochs]
except:
    MNE_EPOCHS_TYPE = []

# centralize configurations
CONFIG = dict(MI_BIASCORRECT=MI_BIASCORRECT, MI_DEMEAN=MI_DEMEAN,
              JOBLIB_CFG=JOBLIB_CFG, MNE_EPOCHS_TYPE=MNE_EPOCHS_TYPE)
